/*
 * $Id: main.c 783309 2016-05-29 20:43:57Z ib-builder $
 *
 * Copyright (c) 2002-2013, Juniper Networks, Inc.
 * All rights reserved.
 */

#include <errno.h>
#include <stdarg.h>
#include <signal.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/ioctl.h>
#include <sys/param.h>
#include <sys/ucred.h>
#include <sys/mount.h>
#include <sys/syslog.h>
#include <sys/ata.h>
#include <sys/stat.h>
#include <sys/dirent.h>
#include <paths.h>
#include <stdlib.h>
#include <stdio.h>
#include <err.h>
#include <ctype.h>

#include "atacmds.h"
#include "smartd.h"
#include "extern.h"
#include "version_gpl.h"

static void Usage (void);

unsigned char driveinfo               = FALSE;
unsigned char checksmart              = FALSE;
unsigned char smartvendorattrib	      = FALSE;
unsigned char generalsmartvalues      = FALSE;
unsigned char smartselftestlog        = FALSE;
unsigned char smarterrorlog           = FALSE;
unsigned char smartdisable            = FALSE;
unsigned char smartenable             = FALSE;
unsigned char smartstatus             = FALSE;
unsigned char smartexeoffimmediate    = FALSE;
unsigned char smartshortselftest      = FALSE;
unsigned char smartextendselftest     = FALSE;
unsigned char smartshortcapselftest   = FALSE;
unsigned char smartextendcapselftest  = FALSE;
unsigned char smartselftestabort      = FALSE;
unsigned char smartautoofflineenable  = FALSE;
unsigned char smartautoofflinedisable = FALSE;
unsigned char formatunit	      = FALSE;
unsigned char issuestandby	      = FALSE;

unsigned int checktime;
unsigned int cmd_standbyinterval = 0;		/* in seconds */
unsigned int cmd_loglevel = 0;			/* Default log level */
unsigned int cmd_summaryperiod = 0;		/* Summary Period */
int smartdpid = -1;
evContext smartd_ctx;

int log_stderr;
int trace_level;

unsigned char smartd_sleep_infinitely = 0;     /* JUNOS */
/*
 * process sighup
 */
static void
smartd_process_sighup (evContext cx, void *uap __unused,
                       const void *tag __unused)
{
    /* Allow another config event to be queued */
    if( 0 > evWaitFor(cx, &smartd_process_sighup, smartd_process_sighup,
                      NULL, NULL)) {
        logging(LOG_EMERG, "%s: evWaitFor failed :%s \n", __FUNCTION__, strerror(errno)); 
    }

    /* Reread config */
    readstandbyconfig();
    readconfigfile(); /* JUNOS */
}

static void
sighup_handler (int signo __unused)
{
    int error = errno;
    
    /* JUNOS begin */
    /* Need to read the config files even if the smartd is 
     * on sabbatical
     */ 
    if (smartd_sleep_infinitely) {
        smartd_process_sighup(smartd_ctx, NULL, NULL);
    /* JUNOS end */	
    } else if (evDo(smartd_ctx, &smartd_process_sighup) < 0) {
        if (errno != ENOENT) {
            logging(LOG_EMERG, "%s: evDo failed :%s \n", __FUNCTION__, strerror(errno)); 
            /* not reached */
        }
    }

    errno = error;
}


/*
 * process sigterm
 */
static void
smartd_process_die_donald_die(int signo __unused)
{
    logging(LOG_DEBUG, "SIGTERM received, exiting");
    exit(0);
}

static int
pid_update (int fd)
{
    char pid_buf[10];


    memset(pid_buf, '\0', sizeof(pid_buf));
    snprintf(pid_buf, sizeof pid_buf, "%d\n", smartdpid);

    if (lseek(fd, 0, SEEK_SET) < 0) {
	logging(LOG_ERR, "%s: lseek: %s", __FUNCTION__, strerror(errno));
	return -1;
    }

    if (write(fd, pid_buf, strlen(pid_buf)) < 0) {
	logging(LOG_ERR, "%s: write: %s", __FUNCTION__, strerror(errno));
	return -1;		/* failure */
    }

    return 0;			/* success */

}

static int
pid_lock (const char *filename)
{
    int fd;

    fd = open(filename, O_CREAT|O_WRONLY, 0644);
    if (fd < 0) {
	logging(LOG_ERR, "error opening %s for writing: %s", filename, strerror(errno));
	return -1;
    }

    if (flock(fd, LOCK_EX|LOCK_NB) < 0) {
        logging(LOG_ERR, "Unable to lock %s: %s", filename, strerror(errno));
	close(fd);
	return -1;
    }

    if (pid_update(fd) < 0) {
        close(fd);
        return -1;
    }

    return fd;			/* success */
}


struct bufcookie {
    char *base;	/* start of buffer */
    int	left;
};

static int
writehook (void *cookie, const char *buf, int len)
{
    struct bufcookie *h;	/* private `handle' */

    h = (struct bufcookie *)cookie;
    if (len > h->left) {
        /* clip in case of wraparound */
        len = h->left;
    }
    if (len > 0) {
        (void)memcpy(h->base, buf, len); /* `write' it. */
        h->base += len;
        h->left -= len;
    }
    return 0;
}

/* Same functionality as vsyslog, uses stderr instead */

static void
vlogging_stderr (const char *fmt, va_list ap)
{
    struct bufcookie fmt_cookie;
    char fmt_cpy[BUFSIZ], ch;
    FILE *fmt_fp;
    int   saved_errno = errno;

    /* Check to see if we can skip expanding the %m */
    if (strstr(fmt, "%m")) {

        /* Create the second stdio hook */
        fmt_cookie.base = fmt_cpy;
        fmt_cookie.left = sizeof(fmt_cpy) - 1;
        fmt_fp = fwopen(&fmt_cookie, writehook);
        if (fmt_fp == NULL) {
            (void)vfprintf(stderr, fmt, ap);
            errno = saved_errno;
            return;
        }

        /* Substitute error message for %m. */
        for ( ; (ch = *fmt); ++fmt)
            if (ch == '%' && fmt[1] == 'm') {
                ++fmt;
                fputs(strerror(saved_errno), fmt_fp);
            } else {
                fputc(ch, fmt_fp);
            }

        /* Null terminate if room */
        fputc(0, fmt_fp);
        fclose(fmt_fp);

        /* Guarantee null termination */
        fmt_cpy[sizeof(fmt_cpy) - 1] = '\0';

        fmt = fmt_cpy;
    }

    (void)vfprintf(stderr, fmt, ap);
    (void)fputc('\n', stderr);
    errno = saved_errno;
}

/*
 * logging()
 *
 * In-place replacement for syslog() library routine.
 * Will log to stderr or to system log.
 */

void
logging (int severity, const char *format, ...) 
{
    va_list ap;

    switch(severity) {
    case LOG_DEBUG:
        if (trace_level < LOG_DEBUG) 
            return;
    case LOG_INFO:
        if (trace_level < LOG_INFO) 
            return;
    default:
        break;
    }

    va_start(ap, format);

    if (log_stderr) {
        vlogging_stderr(format, ap);
    }

    vsyslog(severity, format, ap);

    if (severity == LOG_EMERG)
        abort();

    va_end(ap);
}

void 
Usage (void)
{
    fprintf(stderr, "Non-daemon usage: smartd [-o function] device\n");
    fprintf(stderr, "Read Only Commands:\n");
    fprintf(stderr, "\t\t%s\t\tShow All S.M.A.R.T. Information \n", "oa");
    fprintf(stderr, "\t\t%s\t\tShow General S.M.A.R.T. Attributes\n", "og");
    fprintf(stderr, "\t\t%s\t\tShow Vendor S.M.A.R.T. Attributes\n", "ov");
    fprintf(stderr, "\t\t%s\t\tShow S.M.A.R.T. Drive Error Log \n", "ol");
    fprintf(stderr, "\t\t%s\t\tShow S.M.A.R.T. Drive SelfTest Log \n", "oL");
    fprintf(stderr, "\t\t%s\t\tShow S.M.A.R.T. Drive Info\n", "oi");
    fprintf(stderr, "\t\t%s\t\tCheck S.M.A.R.T. Status\n", "oc");
    fprintf(stderr, "\n");
    fprintf(stderr, "Enable / Disable Commands:\n");
    fprintf(stderr, "\t\t%s\t\tEnable S.M.A.R.T. data collection\n", "oe");
    fprintf(stderr, "\t\t%s\t\tDisable S.M.A.R.T. data collection\n", "od");
    fprintf(stderr, "\t\t%s\t\tEnable S.M.A.R.T. Automatic Offline Test\n", "ot");
    fprintf(stderr, "\t\t%s\t\tDisable S.M.A.R.T. Automatic Offline Test \n", "oT");
    fprintf(stderr, "\n");
    fprintf(stderr, "Test Commands:\n");
    fprintf(stderr, "\t\t%s\t\tExecute Off-line data collection\n", "oO");
    fprintf(stderr, "\t\t%s\t\tExecute Short Self Test\n", "oS");
    fprintf(stderr, "\t\t%s\t\tExecute Short Self Test (Captive Mode)\n", "os");
    fprintf(stderr, "\t\t%s\t\tExecute Extended Self Test \n", "oX");
    fprintf(stderr, "\t\t%s\t\tExecute Extended Self Test (Captive Mode)\n", "ox");
    fprintf(stderr, "\t\t%s\t\tExecute Self Test Abort\n", "oA");
}

static void
query_dev_name(const char *path, char *buf, int len)
{
    int fd;
    struct fiodgname_arg fgn;
    const char *cp;

    fd = open(path, O_RDONLY);
    if (fd >= 0) {
	fgn.buf = buf;
	fgn.len = len;
	if (ioctl(fd, FIODGNAME, &fgn) >= 0) {
	    close(fd);
	    return;
	}
	close(fd);
    }

    /* fallback, assume device is basename */
    cp = strrchr(path, '/');
    if (!cp)
	cp = path;
    else
	cp++;

    strlcpy(buf, cp, len);
}

static int
issamedisk(const char *dev1, const char *dev2)
{
    char realdev1[MAXNAMLEN + 1], realdev2[MAXNAMLEN + 1];
    unsigned int k=0;

    query_dev_name(dev1, realdev1, sizeof(realdev1));
    query_dev_name(dev2, realdev2, sizeof(realdev2));
    
    /* compare non-digit part first */
    while (realdev1[k]) {
	if (isdigit(realdev1[k]))
	    break;
	if (realdev1[k] != realdev2[k])
	    return FALSE;
	k++;
    }

    /* compare digit-only part next */
    while (realdev1[k]) {
	if (!isdigit(realdev1[k]) &&
	    !isdigit(realdev2[k]))
	    break;
	if (realdev1[k] != realdev2[k])
	    return FALSE;
	k++;
    }

    return TRUE;
}

static int 
nondaemonoption(char *devicename, int force)
{
    int i, mntsize;
    int fd;
    struct stat devstat;
    struct statfs *mntbuf;
    char *dname;
    char device[MAXPATHLEN], mntdevice[MAXPATHLEN];

    if (devicename == NULL) {
    fprintf(stderr, "%s: A disk device must be specified\n", SMARTD);
    Usage();
    exit(1);
    }

    if (*devicename != '/') {
        strlcpy(device, _PATH_DEV, sizeof(device));
        strlcat(device, devicename, sizeof(device));
        devicename = device;
    }

    if (stat(devicename, &devstat) != 0 || !(S_ISCHR(devstat.st_mode))) {
    fprintf(stderr, "%s: Must specify a valid device special file\n",
                        SMARTD);
    exit(1);
    }

    if (!force && (smartexeoffimmediate ||  smartshortcapselftest ||
               smartextendcapselftest || formatunit)) {

    if ((mntsize = getmntinfo(&mntbuf, MNT_NOWAIT)) == -1) {
        fprintf(stderr, "%s: Failed to get mount table information\n",
                        SMARTD);
        exit(1);
    }

    for (i = 0; i < mntsize; i++) {
        /*
        if (strcmp(mntbuf[i].f_fstypename, "ufs") != 0)
            continue;
        */
        dname = mntbuf[i].f_mntfromname;
        if (*dname != '/') {
            strlcpy(mntdevice, _PATH_DEV, sizeof(mntdevice));
            strlcat(mntdevice, dname, sizeof(mntdevice));
            dname = mntdevice;
        }
        if (issamedisk (devicename, dname)) {
            fprintf(stderr, "%s: %s is busy. Please unmount %s,"
                 " to use this option.\n", SMARTD, devicename, dname);
            exit(1);
        }
    }
    }

    if ((fd = open(devicename, O_RDONLY)) < 0) {
	fprintf(stderr, "%s: open failed for %s\n", SMARTD, devicename);
	exit(1);
    }

    if (formatunit) {
    	char response[256];

    	fprintf(stderr, "This will erase all data. Are you sure? ");
	scanf("%1s", response);
	if (response[0] != 'y' && response[0] != 'Y') {
    	    fprintf(stderr, "%s: Aborting format command\n", SMARTD);
	    exit(1);
	}
    }

    /* 
     * If disk is in standby and the requested non daemon option needs
     * an ioctl that can bring disk out of standby then we honor it
     * only if force option is set. This is to allow disk to remain in
     * standby as long as possible.
     */
    if (driveinfo || checksmart || generalsmartvalues ||
	smartvendorattrib || smarterrorlog || smartselftestlog) {

	if (ATA_POWERMODE_STANDBY == atacheckpowermode(fd, TRUE)) {
	    if (!force) {
		fprintf(stderr, "%s: %s is in standby. Please retry later..\n",
			SMARTD, devicename);
		exit(1);
	    } else {
		printf("%s: %s is currently in standby\n", SMARTD, devicename);
	    }
	}
    }

    return (ataprintmain(fd));
}

int
main(int argc, char **argv)
{
    char c, option = '\0';
    int ret;
    int pid_fd;
    int dont_daemonize, force, stderr_option;
    int ooption;
    int vflag, gflag, v_as_xml;
    char *device = NULL;
    unsigned long ultmp;
    char *ep =  NULL;

    /* initialize and read args */
    gflag = vflag = pid_fd = dont_daemonize = ooption = force = 0;
    v_as_xml = 0;
    stderr_option = FALSE;
    log_stderr = TRUE; /* Begin with stderr */
    checktime = SMARTD_CHECKTIME;
    trace_level = LOG_INFO; /* Default trace level */

    /* Open log early so we can log argument parse errors. */
    openlog (SMARTD, LOG_PID, LOG_DAEMON);
    
    while ((c = getopt(argc, argv, "J:nNi:d:flo:vX")) != -1) {
	switch (c) {
	case 'o':
	    if (ooption == TRUE) {
	        Usage();
		exit(1);
	    }
	    ooption = TRUE;
	    option = *optarg;
	    switch (option) {
	        case 'A':
		        smartselftestabort	= TRUE;
			break;
		case 'a':
		        driveinfo           	= TRUE;
                        checksmart          	= TRUE;
                        generalsmartvalues  	= TRUE;
                        smartvendorattrib   	= TRUE;
                        smarterrorlog       	= TRUE;
                        smartselftestlog    	= TRUE;
                        break;
	        case 'c':
		        driveinfo 		= TRUE;
                        checksmart		= TRUE;
			break;
		case 'd':
		        smartdisable          	= TRUE;
			break;
		case 'e':
		        smartenable            	= TRUE;
			break;
		case 'F':
		        formatunit            	= TRUE;
			break;
	        case 'g':
		        generalsmartvalues    	= TRUE;
			break;
	        case 'i':
		        driveinfo 		= TRUE;
			break;
	        case 'l':
		        smarterrorlog         	= TRUE;
			break;
	        case 'L':
		        smartselftestlog       	= TRUE;
			break;
		case 'O':
		        smartexeoffimmediate 	= TRUE;
			break;
		case 't':
		        smartautoofflineenable 	= TRUE;
			break;
		case 'S':
		        smartshortselftest 	= TRUE;
			break;
		case 's':
		        smartshortcapselftest 	= TRUE;
			break;
		case 'T':
		        smartautoofflinedisable = TRUE;
			break;
		case 'v':
		        driveinfo               = TRUE;   /* JUNOS */
		        smartvendorattrib	= TRUE;
			break;
		case 'X':
		        smartextendselftest  	= TRUE;
			break;
		case 'x':
		        smartextendcapselftest 	= TRUE;
			break;
		default:		/* unsupported */
			Usage();
			exit(1);
	    }
	    break;
	case 'l':		/* syslog to stderr as well */
	    stderr_option = TRUE;
	    break;
	case 'd':		/* set debugging level */
	    trace_level = atoi(optarg);
	    break;
	case 'n':
	case 'N':		/* Don't daemonize */
	    dont_daemonize = TRUE;
	    break;
	case 'f':		/* force */
	    force = TRUE;
	    break;
	case 'i':		/* device polling time */
            ultmp = strtoul(optarg, &ep, 0);
            if (*ep || ep == optarg || ultmp > (UINT_MAX / 60) || !ultmp) {
                logging(LOG_ERR, "Invalid interval value : '%s'\n", optarg); 
                exit(1);
            }
            checktime = (unsigned int) ultmp * 60;      /* In seconds */
	    break;
	case 'J':
	    /* JNX Extensions */
	    switch (*optarg) {
	    case 'l':
		if (optind == argc) {
		    logging(LOG_ERR,
			"Expected log level: '-J%s <level>' \n", optarg); 
		    exit(1);
		}
		ultmp = strtoul(argv[optind++], &ep, 0);
		if (*ep || ep == optarg || ultmp > LOG_DEBUG || !ultmp) {
		    logging(LOG_ERR, "Invalid log level value \n");
		    exit(1);
		}
		cmd_loglevel = ultmp;
		break;
	    case 't':	/* standby interval, only relevant in daemon mode 
		         * should be only used for testing purpose */
		if (optind == argc) {
		    logging(LOG_ERR,
			"Expected standby interval: '-J%s <time>' \n", optarg); 
		    exit(1);
		}
		ultmp = strtoul(argv[optind++], &ep, 0);
		if (*ep || ep == optarg || !ultmp) {
		    logging(LOG_ERR, "Invalid standby interval value \n");
		    exit(1);
		}
		cmd_standbyinterval = (unsigned int) ultmp; /* In seconds */
		break;
	    case 'T':	/* Time after which summary is printed and
			 * accumulated times are reset */
		if (optind == argc) {
		    logging(LOG_ERR,
			"Expected summary time: '-J%s <time>' \n", optarg); 
		    exit(1);
		}
		ultmp = strtoul(argv[optind++], &ep, 0);
		if (*ep || ep == optarg || !ultmp ||
		    ((UINT_MAX - ultmp) < SMARTD_STANDBY_SUMMARY_OFFSET)) {
		    logging(LOG_ERR, "Invalid standby interval value \n");
		    exit(1);
		}
		cmd_summaryperiod = (unsigned int) ultmp; /* In seconds */
		break;
	    case 'S':/* issue standby immediate */
		ooption = TRUE;
		issuestandby = TRUE;
		break;
	    default:		/* unsupported */
		logging(LOG_ERR, "usage: %s [-lvN] [-i interval] "
		    "[-d debug_level] [-o function] [device]\n",
		    SMARTD);
		exit(1);
	    }
	    break;
	case 'v':		/* print version */
	    vflag++;
	    break;
	case 'X':		/* when printing version use XML */
	    v_as_xml++;
	    break;
	default:		/* unsupported */
	    logging(LOG_ERR, "usage: %s [-lvN] [-i interval] [-d debug_level] "
		    "[-o function] [-J function] [device]\n", SMARTD);
	    exit(1);
	}
    }

    argc -= optind;
    device = argv[optind];
    argv += optind;

    smartdpid = getpid();

    /*
     * process version flag
     */
    if (vflag) {
	printf ("%s\n", swversion_what());
	if (vflag > 1) {
	    printf ("    %s\n", swversion_build_info());
	}
	exit(0);
    }

    /* Make sure we are running as root */
    if (geteuid()) {
	logging(LOG_ERR, "%s: must be run as root", SMARTD);
	exit(1);
    }

    /* Create event library context */
    if (evCreate(&smartd_ctx) < 0) {
	logging(LOG_ERR, "%s: could not create event library context:%s\n",
		__FUNCTION__, strerror(errno));
	exit(1);
    }

    setlogmask(LOG_UPTO(trace_level));

    /* Setup signal handlers */

    signal(SIGHUP, sighup_handler);
    if( 0 > evWaitFor(smartd_ctx, &smartd_process_sighup, smartd_process_sighup,
                      0, 0)) {
        logging(LOG_EMERG, "%s: evWaitFor failed :%s \n", __FUNCTION__, strerror(errno)); 
        /* not reached */
    }

    signal(SIGTERM, smartd_process_die_donald_die);


    if (ooption) {
        syslog(LOG_INFO, "%s executing option -o%c for device %s, pid %d", 
			SMARTD, option, device, smartdpid);
	ret = nondaemonoption(device, force);
	exit(ret);
    }

    /*
     * see if another one is running
     */
    pid_fd = pid_lock(PATH_SMARTD_PIDFILE);
    if (pid_fd < 0) {
        logging(LOG_ERR, "Is another copy of this program running?");
	exit(1);
    }

    /*
     * detach from the controlling terminal and become a daemon
     */
    if (!dont_daemonize && (daemon(0, stderr_option) < 0)) {
	logging(LOG_EMERG, "%s: daemon: %s, aborting", __FUNCTION__, strerror(errno));
    }

    /*
     * update the pid file with our new pid
     */
    if (pid_update(pid_fd) < 0) {
	logging(LOG_ERR, "%s: pid_update failed", __FUNCTION__);
	exit (1);
    }

    /* cut over to syslog from stderr */
    log_stderr = stderr_option;

    logging(LOG_INFO,"%s starting, pid %d", SMARTD, smartdpid);

    /*
     * transfer control to the main loop handler
     */
    ret = daemonloop(device);

    /*
     * we should never reach here
     */
    if (ret == ENOENT) {
        logging(LOG_WARNING, "Exiting, no devices to poll.\n");
    } else {
        logging(LOG_ERR, "Unexpected exit: %d\n", ret);
    }
    exit(ret);
}
